// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en
import { randomBytes, createCipheriv, createDecipheriv } from 'crypto';
import AESKey from './aesKey';
import createHmac from 'create-hmac';
import Utils from './utils';

const sizes = { key: 16, iv: 16, hash: 32 };

export default class AESCipher {
    /** 
     * @param {AESKey} key
     * @param {Uint8Array|string} data
     **/
    static encrypt(key, data) {
        var iv = randomBytes(sizes.iv);
        const cipher = createCipheriv('aes-128-cbc', key.key, iv);

        var buffer = typeof data === 'string' ? Buffer.from(data) : data;
        var body = Buffer.concat([iv, cipher.update(buffer), cipher.final()]);
        var mac = hmac(key.mac, body);

        return Buffer.concat([body, mac]);
    }

    /**
     * @param {AESKey} key
     * @param {string | Uint8Array} data
     */
    static encryptStr(key, data) {
        return Buffer.from(AESCipher.encrypt(key, data)).toString('base64');
    }

    /** 
     * @param {AESKey} key
     * @param {Uint8Array|string} data
     **/
    static decrypt(key, data) {
        var buffer = typeof data === 'string' ? Buffer.from(data, 'base64') : data;
        var hashb = buffer.slice(-sizes.hash);
        buffer = buffer.slice(0, -sizes.hash);

        var hasha = hmac(key.mac, buffer);
        if (!Utils.safeEqual(hasha, hashb))
            throw new Error("Invalid signature");

        var iv = buffer.slice(0, sizes.iv);
        const decipher = createDecipheriv('aes-128-cbc', key.key, iv);
        
        return Buffer.concat([decipher.update(buffer.slice(sizes.iv)), decipher.final()]);
    }

    /**
     * @param {AESKey} key
     * @param {string | Uint8Array} data
     */
    static decryptStr(key, data) {
        return Buffer.from(AESCipher.decrypt(key, data)).toString('utf8');
    }
}

/**
 * @param {Uint8Array} key
 * @param {Uint8Array} data
 */
function hmac(key, data) {
    return createHmac('sha256', Buffer.from(key)).update(data).digest();
}