// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import { randomBytes } from 'crypto';
import { eddsa } from 'elliptic';
var ec = new eddsa('ed25519');

export default class ed25519Key {
    get pub() {
        return new ed25519Key(null, this._pub);
    }

    /**
     * @param {eddsa.KeyPair} [pub]
     * @param {eddsa.KeyPair} [prv]
     * @private
     */
    constructor(prv, pub) {
        /** @private */
        this._prv = prv;
        /** @private */
        this._pub = pub;
    }

    /**@returns {ed25519Key} */
    static generate() {
        const prv = ec.keyFromSecret(randomBytes(32));
        const pub = ec.keyFromPublic(prv.getPublic())

        return new ed25519Key(prv, pub);
    }

    /**
     * @param {Uint8Array|string} key
     * @returns {ed25519Key}
     */
    static public(key) {
        const buff = typeof key === 'string' ?  Buffer.from(key, 'base64') : key;
        const pub = ec.keyFromPublic(Array.from(buff));

        return new ed25519Key(null, pub);
    }

    /**
     * @param {Uint8Array|string} key
     * @returns {ed25519Key}
     */
    static private(key, noPublic = false) {
        const buff = typeof key === 'string' ? Buffer.from(key, 'base64') : Buffer.from(key);
        const prv = ec.keyFromSecret(buff);
        const pub = noPublic ? null : ec.keyFromPublic(prv.getPublic());

        return new ed25519Key(prv, pub);
    }

    /**
     * @param {Uint8Array|string} message 
     * @returns {Uint8Array}
     */
    sign(message) {
        if (!this._prv) throw new Error("private key is required");
        if (!message || !message.length) throw new Error("message is required");

        const msg = typeof message === 'string' ? Buffer.from(message, 'utf-8') : Buffer.from(message);

        return Buffer.from(this._prv.sign(msg).toBytes());
    }

    /**
     * @param {Uint8Array|string} message
     * @param {Uint8Array} signature
     */
    verify(message, signature) {
        if (!this._pub) throw new Error("public key is required");
        if (!message || !message.length) throw new Error("message is required");
        if (!signature || signature.length != 64) return false;

        const msg = typeof message === 'string' ? Buffer.from(message, 'utf-8') : Buffer.from(message);

        return this._pub.verify(msg, Array.from(signature));
    }

    toString() {
        return Buffer.from(this.toBuffer()).toString('base64');
    }

    toBuffer() {
        const buffer = new Uint8Array(64);
        
        if (this._prv) buffer.set(this._prv.secret(), 0);
        if (this._pub) buffer.set(this._pub.getPublic(), 32);
        
        return buffer;
    }

    /**
     * @param {Uint8Array|string} data
     */
    static from(data) {
        if (!data || !data.length) throw new Error("data is required");

        const buff = typeof data === 'string' ? Buffer.from(data, 'base64') : Buffer.from(data);
        
        if (buff.length !== 64) throw new Error("data in invalid format");

        let prv = null;
        if (!buff.slice(0, 32).every(num => num == 0))
            prv = ec.keyFromSecret(buff.slice(0, 32));
        
        let pub = null;
        if (!buff.slice(32).every(num => num == 0))
            pub = ec.keyFromPublic(Array.from(buff.slice(32)));

        return new ed25519Key(prv, pub);
    }
}
