import BN from 'bn.js';
import assert from 'assert';
import { randomBytes } from 'crypto';
import { eddsa as EdDSA } from 'elliptic';
import { getBig, getBN } from '../src/bnInput';
import SecretShare from '../src/secretShare';
import Hash from '../src/hash';

describe('tssEdDsa', function () {
    it('Should sign in a threshold', function tssEdDsa() {
        const ec = new EdDSA('ed25519');
        const curve = ec.curve;
        const g = curve.g;
        g.precompute(curve.n.bitLength() + 1);

        const hash = Hash.sha512Buffer;
        const key = ec.keyFromSecret(randomBytes(32));

        /** @type {BN} */
        const s = key.priv();
        const A = g.mul(s);

        const message = Buffer.from("hi there!!!", "base64");
        const threshold = 15;
        const ids = [...Array(20)].map((_, id) => getBN(id));

        const ris = [...Array(ids.length)].map(_ => new BN(randomBytes(32), 'le').mod(curve.n));
        const sis = share(s, ids, threshold, curve.n);

        const Ris = ris.map(ri => g.mul(ri));
        const R = Ris.reduce((Rsum, Ri) => Rsum.add(Ri));

        const hBuffer = hash(Buffer.concat([encodePoint(R), encodePoint(A), message]));
        const h = new BN(hBuffer, 'le').mod(curve.n);
        const Sis = ids.map(id => getLi(id, ids, curve.n))
            .map((li, i) => h.mul(sis[i]).mul(li).add(ris[i]).mod(curve.n));

        const S = Sis.reduce((Ssum, Si) => Ssum.add(Si).mod(curve.n));
        const signature = Buffer.concat([encodePoint(R), S.toBuffer('le')]);

        const keyPub = ec.keyFromPublic(key.getPublic());
        const sw = keyPub.verify(message, signature.toString('hex'));
        
        assert.ok(sw)
    });
});


/** @param {import("elliptic").curve.base.BasePoint} point */
function encodePoint(point) {
    const encodingLength = 32;
    const enc = point.getY().toArray('le', encodingLength);
    enc[encodingLength - 1] |= point.getX().isOdd() ? 0x80 : 0;
    return Buffer.from(enc);
}

/**
 * @param {BN} id
 * @param {BN[]} ids
 * @param {BN} mod
 */
function getLi(id, ids, mod) {
    const bigId = getBig(id);
    const bigIds = ids.map(id => getBig(id));
    return getBN(SecretShare.getLi(bigId, bigIds, getBig(mod)));
}

/**
 * @param {BN} secret
 * @param {BN[]} ids
 * @param {number} threshold
 * @param {BN} mod
 */
function share(secret, ids, threshold, mod) {
    const x = getBig(secret);
    const n = getBig(mod);
    const nums = ids.map(id => getBig(id));
    const [, shrs] = SecretShare.shareFromIds(x, nums, threshold, n);
    return shrs.map(shr => getBN(shr));
}
